from __future__ import annotations
import numpy as np
from typing import TYPE_CHECKING
from mpi4py import MPI
import sys

if TYPE_CHECKING:
    from .partition import Partition
    from .spherical_partition import S2Partition

ParticleDataT = dict[str, np.ndarray]


def distribute_dataset_by_home(
    partition: Partition | S2Partition,
    data: ParticleDataT,
    home_idx: np.ndarray,
    *,
    verbose: int = 0,
    verify_count: bool = True,
    all2all_iterations: int = 1,
) -> ParticleDataT:
    total_to_send = len(home_idx)
    nperiteration = total_to_send // all2all_iterations
    data_new_list: list[ParticleDataT] = []

    # Some general assertions that every rank has valid data
    keys = list(data.keys())
    keys_0 = partition.comm.bcast(keys, root=0)
    assert len(keys) == len(keys_0)
    assert all(k in keys_0 for k in keys)
    dtype_string = "".join(data[k].dtype.char for k in keys_0)
    dtype_string_0 = partition.comm.bcast(dtype_string, root=0)
    assert dtype_string == dtype_string_0

    for i in range(all2all_iterations):
        start_idx = i * nperiteration
        end_idx = (
            (i + 1) * nperiteration if i < all2all_iterations - 1 else total_to_send
        )
        if partition.rank == 0 and verbose > 0:
            print(f"  - Distributing particles iteration {i + 1}/{all2all_iterations}")
        _data = {k: v[start_idx:end_idx] for k, v in data.items()}
        _home_idx = home_idx[start_idx:end_idx]
        _data = _distribute_dataset_by_home(
            partition,
            _data,
            _home_idx,
            keys_0,
            verbose=verbose,
            verify_count=verify_count,
        )
        data_new_list.append(_data)
    # concatenate the data
    data_new = {k: np.concatenate([d[k] for d in data_new_list]) for k in data.keys()}
    return data_new


def _distribute_dataset_by_home(
    partition: Partition | S2Partition,
    data: ParticleDataT,
    home_idx: np.ndarray,
    keys: list[str],
    *,
    verbose: int = 0,
    verify_count: bool = True,
) -> ParticleDataT:
    total_to_send = len(home_idx)
    for d in data.values():
        assert len(d) == total_to_send, "All data arrays must have the same length"

    # sort by rank
    s = np.argsort(home_idx)
    home_idx = home_idx[s]

    # offsets and counts
    send_displacements = np.searchsorted(home_idx, np.arange(partition.nranks))
    send_displacements = send_displacements.astype(np.int32)
    send_counts = np.append(send_displacements[1:], total_to_send) - send_displacements
    send_counts = send_counts.astype(np.int32)

    # announce to each rank how many objects will be sent
    recv_counts = np.empty_like(send_counts)
    partition.comm.Alltoall(send_counts, recv_counts)
    recv_displacements = np.insert(np.cumsum(recv_counts)[:-1], 0, 0)

    # number of objects that this rank will receive
    total_to_receive = np.sum(recv_counts)

    # debug message
    if verbose > 1:
        for i in range(partition.nranks):
            if partition.rank == i:
                print(f"Distribute Debug Rank {i}")
                print(f" - rank has {total_to_send} particles")
                print(f" - rank receives {total_to_receive} particles")
                print(f" - send_counts:        {send_counts}")
                print(f" - send_displacements: {send_displacements}")
                print(f" - recv_counts:        {recv_counts}")
                print(f" - recv_displacements: {recv_displacements}")
                print("", flush=True)
            partition.comm.Barrier()

    # send data all-to-all, each array individually
    data_new = {k: np.empty(total_to_receive, dtype=data[k].dtype) for k in data.keys()}

    for k in keys:
        d = data[k][s]
        s_msg = [d, (send_counts, send_displacements), d.dtype.char]
        r_msg = [data_new[k], (recv_counts, recv_displacements), d.dtype.char]
        partition.comm.Alltoallv(s_msg, r_msg)

    if verify_count:
        key0 = keys[0]
        local_counts = np.array([len(data[key0]), len(data_new[key0])], dtype=np.int64)
        global_counts = np.empty_like(local_counts)
        partition.comm.Reduce(local_counts, global_counts, op=MPI.SUM, root=0)
        if partition.rank == 0 and global_counts[0] != global_counts[1]:
            print(
                f"Error in distribute: particle count during distribute was not "
                f"maintained ({global_counts[0]} -> {global_counts[1]})",
                file=sys.stderr,
                flush=True,
            )
            partition.comm.Abort()

    return data_new



import sys
from typing import List, Union

import numpy as np

from .partition import Partition
from ._send_home import distribute_dataset_by_home

ParticleDataT = dict[str, np.ndarray]


def distribute(
    partition: Partition,
    box_size: float,
    data: ParticleDataT,
    coord_keys: List[str],
    *,
    verbose: Union[bool, int] = False,
    verify_count: bool = True,
    all2all_iterations: int = 1,
) -> ParticleDataT:
    """Distribute data among MPI ranks according to data position and volume partition

    The position of each TreeData element is given by the data columns
    specified with `coord_keys`.

    Parameters
    ----------

    partition:
        The MPI partition defining which rank should own which subvolume of the
        data

    box_size:
        The size of the full simulation volume

    data:
        The treenode / coretree data that should be distributed

    coord_keys:
        The columns in `data` that define the position of the object

    verbose:
        If True, print summary statistics of the distribute. If > 1, print
        statistics of each rank (i.e. how much data each rank sends to every
        other rank).

    verify_count:
        If True, make sure that total number of objects is conserved

    all2all_iterations:
        The number of iterations to use for the all-to-all communication.
        This is useful for large datasets, where MPI_Alltoallv may fail

    Returns
    -------
    data: ParticleDataT
        The distributed particle data (i.e. the data that this rank owns)

    """
    assert len(coord_keys) == partition.dimensions

    # get some MPI and partition parameters
    nranks = partition.nranks
    if nranks == 1:
        return data

    # rank = partition.rank
    comm = partition.comm
    dimensions = partition.dimensions
    ranklist = np.array(partition.ranklist)
    extent = box_size * np.array(partition.extent)

    # count number of particles we have
    total_to_send = len(data[coord_keys[0]])

    if total_to_send > 0:
        # Check validity of coordinates
        for i in range(dimensions):
            _x = data[coord_keys[i]]
            _min = _x.min()
            _max = _x.max()
            if _min < 0 or _max > box_size:
                print(
                    f"Error in distribute: position {coord_keys[i]} out of range: [{_min}, {_max}]",
                    file=sys.stderr,
                    flush=True,
                )
                comm.Abort()

        # Find home of each particle
        idx = np.array(
            [data[coord_keys[i]] / extent[i] for i in range(dimensions)]
        ).astype(np.int32)
        idx = np.clip(idx, 0, np.array(partition.decomposition)[:, np.newaxis] - 1)
        home_idx = ranklist[tuple(idx)]
    else:
        # there are no particles on this rank
        home_idx = np.empty(0, dtype=np.int32)

    data_new = distribute_dataset_by_home(
        partition,
        data,
        home_idx=home_idx,
        verbose=verbose,
        verify_count=verify_count,
        all2all_iterations=all2all_iterations,
    )

    return data_new



from typing import Union
import numpy as np

from .s2_partition import S2Partition
from .._send_home import distribute_dataset_by_home

ParticleDataT = dict[str, np.ndarray]


def s2_distribute(
    partition: S2Partition,
    data: ParticleDataT,
    *,
    theta_key: str = "theta",
    phi_key: str = "phi",
    verbose: Union[bool, int] = False,
    verify_count: bool = True,
    validate_home: bool = False,
    all2all_iterations: int = 1,
) -> ParticleDataT:
    """Distribute particles among MPI ranks according to the S2 partition.

    Parameters
    ----------
    partition:
        The S2 partition to use for the distribution.

    data:
        The particle data to distribute, as a collection of 1-dimensional arrays.
        Each array must have the same length (number of particles) and the map needs
        to contain at least the keys `theta_key` and `phi_key`.

    theta_key:
        The key in `data` that contains the particle theta coordinates (latitude),
        in the range [0, pi].

    phi_key:
        The key in `data` that contains the particle phi coordinates (longitude),
        in the range [0, 2*pi].

    verbose:
        If True, print summary statistics of the distribute. If > 1, print
        statistics of each rank (i.e. how much data each rank sends to every
        other rank).

    verify_count:
        If True, make sure that total number of objects is conserved.

    validate_home:
        If True, validate that each rank indeed owns the particles that it was sent.

    all2all_iterations:
        The number of iterations to use for the all-to-all communication.
        This is useful for large datasets, where MPI_Alltoallv may fail.

    Returns
    -------
    data: ParticleDataT
        The distributed particle data (i.e. the data that this rank owns)

    """

    # verify data is normalized
    assert np.all(data[theta_key] >= 0)
    assert np.all(data[theta_key] <= np.pi)
    assert np.all(data[phi_key] >= 0)
    assert np.all(data[phi_key] < 2 * np.pi)

    # ring idx: 0=cap, 1=first ring, 2=second ring, etc.
    if partition.equal_area:
        ring_idx = np.digitize(data[theta_key], partition.ring_thetas)
    else:  # equal theta
        if partition.nranks == 2:
            ring_idx = (data[theta_key] > partition.theta_cap).astype(np.int32)
        else:
            assert partition.ring_dtheta is not None
            ring_idx = (
                (data[theta_key] - partition.theta_cap) // partition.ring_dtheta
            ).astype(np.int32) + 1
            ring_idx = np.clip(ring_idx, 0, len(partition.ring_segments) + 1)
            ring_idx[data[theta_key] == np.pi] -= 1  # handle cases where theta == pi

    phi_idx = np.zeros_like(ring_idx, dtype=np.int32)
    mask_is_on_ring = (ring_idx > 0) & (ring_idx <= len(partition.ring_segments))
    phi_idx[mask_is_on_ring] = (
        data[phi_key][mask_is_on_ring]
        / (2 * np.pi)
        * partition.ring_segments[ring_idx[mask_is_on_ring] - 1]
    ).astype(np.int32)

    # rank index where each ring starts
    ring_start_idx = np.zeros(len(partition.ring_segments) + 2, dtype=np.int32)
    ring_start_idx[1] = 1
    ring_start_idx[2:] = np.cumsum(partition.ring_segments) + 1

    # rank index of each particle
    home_idx = ring_start_idx[ring_idx] + phi_idx

    assert np.all(home_idx >= 0)
    assert np.all(home_idx < partition.nranks)

    data_new = distribute_dataset_by_home(
        partition,
        data,
        home_idx=home_idx,
        verbose=verbose,
        verify_count=verify_count,
        all2all_iterations=all2all_iterations,
    )

    if validate_home:
        assert np.all(data_new[theta_key] >= partition.theta_extent[0])
        if partition.theta_extent[1] < np.pi:
            assert np.all(data_new[theta_key] < partition.theta_extent[1])
        else:
            # bottom cap, we allow theta == pi
            assert np.all(data_new[theta_key] <= partition.theta_extent[1])
        assert np.all(data_new[phi_key] >= partition.phi_extent[0])
        assert np.all(data_new[phi_key] < partition.phi_extent[1])

    return data_new
